// SPDX-License-Identifier: GPL-2.0-only
/* Copyright (c) 2021 Universidad de Palermo - Argentina - https://palermo.edu

 * TCP Palermo+: Sender Side Congestion Control and Buffer Bloating Prevention
 * Version: 2.06
 * Authors:
 *		Alejandro Popovsky: Universidad de Palermo, project director
 *    Gustavo Muzzillo:   Universidad de Palermo, research collaborator
 * 
 */

#include <linux/mm.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/inet_diag.h>
#include <net/tcp.h>
#include <linux/sysctl.h>
#include <linux/prefetch.h>
#include <linux/slab.h>
#include <linux/kernel.h>

#define PALDEBUG 1

char* pal_cstateTable[] = {
	"nonCwndLim",		/* 0: Not congestion Window limited */
	"init",				/* 1: Initial state (no last_ackrate yet known */
	"regular",			/* 2: regular search for capacity */
	"gentleFight",		/* 3: respect for others sharing the bottleneck */
	"dogFight",			/* 4: regain capacity from aggressive competitor */
	"queueRecover1",	/* 5: recover from opt rate overestimation */
	"queueRecover2"	/* 6: recover from opt window too big */
};

// Note: blindQueueRecover states are used when cwnd seems larger than base BDP,
// but still being cwnd_limited. If being app_limited or recv_win_limited then
// cwnd is decreased by another mechanism in main tcp logic: rfc286: cong window
// validation.
#define nonCwndLimState 0
#define initState 1
#define regularState 2
#define gentleFightState 3
#define dogFightState 4
#define queueRecover1State 5
#define queueRecover2State 6

// module params
static int win_regular_growth8 __read_mostly = (400<<8)/1000;
module_param(win_regular_growth8 ,int,0644);
MODULE_PARM_DESC(win_regular_growth8, "window growth % after normal rate increase");

static int win_regular_decrease8 __read_mostly = -(285<<8)/1000;
module_param(win_regular_decrease8 ,int,0644);
MODULE_PARM_DESC(win_regular_decrease8,"window decrease when needed");

static int win_small_growth8 __read_mostly = (200<<8)/1000;
static int win_small_decrease8 __read_mostly = -(166<<8)/1000;
int set_small_growth8(const char *val, const struct kernel_param *kp) {
	int n,ret; ret = kstrtoint(val,10,&n);
	if( ret!=0 || n>512 || n<=0 ) return -EINVAL;
	win_small_decrease8 = -((n<<8)/(256+n));
	return param_set_int(val, kp);
}
int set_small_decrease8(const char *val, const struct kernel_param *kp) {
	int n,ret; ret = kstrtoint(val,10,&n);
	if( ret!=0 || n>512 || n>=0 ) return -EINVAL;
	win_small_growth8 = ((-n)<<8)/(256+n);
	return param_set_int(val, kp);
}
static const struct kernel_param_ops win_small_decrease8_ops = {
	.set = set_small_decrease8,
	.get = param_get_int,
};
static const struct kernel_param_ops win_small_growth8_ops = {
	.set = set_small_growth8,
	.get = param_get_int,
};

module_param_cb(win_small_growth8,&win_small_growth8_ops,&win_small_growth8,0644);
MODULE_PARM_DESC(win_small_growth8,"window growth with caution");

module_param_cb(win_small_decrease8,&win_small_decrease8_ops,&win_small_decrease8,0644);
MODULE_PARM_DESC(win_small_decrease8,"window decrease when needed");

static int x_thresholdK8 __read_mostly = (600<<8)/1000;
module_param(x_thresholdK8,int,0644);
MODULE_PARM_DESC(x_thresholdK8,"relative rate/win limit to enter regular window growth");

static int x_thresholdL8 __read_mostly = (1300<<8)/1000;
module_param(x_thresholdL8,int,0644);
MODULE_PARM_DESC(x_thresholdL8,"relative rate/win limit to enter cautious window growth");

static int opt_check __read_mostly = 1;
module_param(opt_check,int,0644);
MODULE_PARM_DESC(opt_check,"enable optimum values check");

static int ss_fast __read_mostly = 1;
module_param(ss_fast,int,0644);
MODULE_PARM_DESC(ss_fast,"regular growth during slow start");

static int x_thresholdLOW8 __read_mostly = (150<<8)/1000;
module_param(x_thresholdLOW8,int,0644);
MODULE_PARM_DESC(x_thresholdLOW8,"relative rate/win smaller than this is considered 0");

static int x_REP = 3;
module_param(x_REP,int,0644);
MODULE_PARM_DESC(x_REP,"repetitions needed for relative rate/win to assume overestimation");

static int rate_thresholdAPROX8 __read_mostly = (906<<8)/1000; 
module_param(rate_thresholdAPROX8,int,0644);
MODULE_PARM_DESC(rate_thresholdAPROX8,"% rate/optRate tolerance (92.5%)");

static int rate_thresholdDRS8 __read_mostly = (200<<8)/1000;
module_param(rate_thresholdDRS8,int,0644);
MODULE_PARM_DESC(rate_thresholdDRS8,"% rate/optRate decrease to enter dogFight");

static int useX __read_mostly = 1;
module_param(useX,int,0644);
MODULE_PARM_DESC(useX,"calculate share of capacity based on rate useX=1 or based on rtt useX=0");

static int derivRef __read_mostly = 1;
module_param(derivRef,int,0644);
MODULE_PARM_DESC(useX,"calculate derivative referred to avg value derivRef=1 or referred to last value derivRef=0");

static int guardWindows __read_mostly = 1;
module_param(guardWindows,int,0644);
MODULE_PARM_DESC(useX,"Number of full guard windows to wait before calculating Xa or Ya");

/* Parameters from receiver side palermo congestion control, not used in sender side version */
/* int afterfastrec_backoff __read_mostly = 1; postpone DRS after fast recovery */
/* int assume_prr_peer __read_mostly = 1; assume sender will do rate halving on fast recovery */
/* int palermo_rtt_tau __read_mostly = 0; 0<=tau<=7 rtt filter convergence speed. 0 fastest, 7 slowest */
/* int dg_max __read_mostly = 3; limit for data generation detection */
/* int rwin_guard __read_mostly = 2; limit for data generation detection */
/* int rate_rise_factor __read_mostly = 57; rate rise time expressed in rtt's x 57 */


/* TCP palermo structure */
struct palermo {
	u32 last_ack;			/* last ack when congestion window was modified */
	u64 last_ack_us;		/* system time during last xa measure*/
	u32 last_ackrate; 	/* transmision rate during previous xa mesure*/
	u32 last_srtt_us;		/* last rtt during previous ya measure */
	u32 last_win;			/* congestion window on last checkpoint. Expressed in tp->mss_cache's */
	u32 prev_last_win;	/* congestion window 2 checkpoints ago. Expressed in tp->mss_cache's */
	s32 winvar;				/* current increase in window. Expressed in tp->mss_cache's. Gradually applied during window. */
	u32 opt_win;	   	/* Optimal window size */
	u32 opt_rate;	   	/* Optimal ack rate */	
	u32 opt_srtt_us;		/* Optimal rtt */
	u8  state;				/* state depicting current connection behavior */
	u8  osc;
	s8  ls;
	u8  guard;				/* wait guard before measure effects of window variation */
};

/* circular arithmetic 32 */
inline u32 max32(u32 a,u32 b) { return (((s32)(a-b))>0)?a:b; }
inline u32 min32(u32 a,u32 b) { return (((s32)(a-b))>0)?b:a; }
inline s32 dif32(u32 a,u32 b) { return (s32)(a-b); } /* ok */
inline char gt32(u32 a,u32 b) { return ((s32)(a-b))>0; }
inline char gte32(u32 a,u32 b) { return ((s32)(a-b))>=0; }

#define Z8UNINIT 1128374619
#define Z8_DEFINED(za) (za!=Z8UNINIT)
#define Z8_UNDEFINED(za) (za==Z8UNINIT)

inline int changeRatio8(s32 currentR, s32 lastR, s32 currentW, s32 lastW) {
	/* Work with s64 because before division we may have very large numbers.
		Overflow check example: currentRate=12500000 (equivalent to 100gbps)
		deltaRate=10000000, lastWin = 900MBytes.
		=> deltaRate*rcvd_lastWin*256 = aprox 2^61 < 2^63 ok */ 
	s64 x8  = (s64)(currentR-lastR); // 8bits for fractional part instead of floats
	s64 div = (s64)(currentW-lastW);
	if( derivRef ) {
		x8  *= (s64)currentW+(s64)lastW;
		div *= (s64)currentR+(s64)lastR;
	} else {
		x8  *= (s64)lastW;
		div *= (s64)lastR;
	}
	x8 <<= 8;
	if( !div ) return Z8UNINIT; /* ths can only happen for an unchanged window */
	x8 /= div;
	return (int)x8;
}

/*
 * @tcp_palermo_init
 * This function initializes fields used in TCP Palermo+,
 * it is called after the initial SYN, so the sequence numbers
 * are correct but new passive connections we have no
 * information about RTTmin at this time so we simply set it to
 * TCP_PALERMO_INIT_RTT. This value was chosen to be too conservative
 * since in this way we're sure it will be updated in a consistent
 * way as soon as possible. It will reasonably happen within the first
 * RTT period of the connection lifetime.
 */
static void tcp_palermo_init(struct sock *sk)
{	struct tcp_sock *tp = tcp_sk(sk);
	struct palermo *pal = inet_csk_ca(sk);
#ifdef PALDEBUG
	struct inet_sock *inet = inet_sk(sk);
#endif
	pal->last_ack = 0;
	pal->last_ack_us = 0;
	pal->last_ackrate = 0;
	pal->last_srtt_us = (u32)-1;
	pal->last_win = 0;
	pal->prev_last_win = 0;
	pal->winvar = tp->snd_cwnd; /* slow start, duplicste window during current window transmission */
	pal->opt_win = 0;
	pal->opt_rate = 0;
	pal->opt_srtt_us = (u32)-1;
   pal->state = initState;
	pal->osc = 0;
   pal->ls=0;
	pal->guard = 0;
#ifdef PALDEBUG
	pr_info("tcp_palermo_init() step=init conn=%u-%pI4-%u",
			htons(inet->inet_sport),
			&inet->inet_saddr,
			htons(inet->inet_dport));
#endif
}

static inline s32 SCALE8(s32 x,s32 y) { return (x*y)>>8; }

static inline void setWinVar(struct tcp_sock *tp,struct palermo* pal,s32 x) {
	 pal->winvar = SCALE8(tp->snd_cwnd,x);
	 pal->osc = 0;
}

/* pure reno winVar */
static inline void setWinVarRegular(struct tcp_sock *tp,struct palermo* pal) {
	pal->winvar = tcp_in_slow_start(tp) ?
			/*slow start     */ ( ss_fast ? tp->snd_cwnd : SCALE8(tp->snd_cwnd,win_regular_growth8) ) :
		  	/*cong avoidance */  2;
			/* only 1 per rtt is too small to make Xa measurements, when restarting a large win.
			 * But larger windows may be dangerous when restarting after fast recovery */
	pal->osc = 0;
}

static inline void setWinVarBounce(struct tcp_sock *tp,struct palermo* pal) {
	 pal->winvar = - pal->winvar;
	 pal->osc = 0;
}

static inline void setWinVarOscilate(struct tcp_sock *tp,struct palermo* pal) {
	if( ! opt_check ) {
		pal->winvar = 0; /* the window will not vary. It is disabled when !opt_check */
		pal->osc = 0;
		return;
	}
	if( pal->osc ) {
		/* At this point we should be exactly at tp->snd_cwnd = pal->last_win + pal->winvar
		 * but sometimes we fall short in reching this value. So to prevent slow drifts
		 * during oscillate, we make sure this is true */
		/* tp->snd_cwnd = pal->last_win + pal->winvar */
		/* from here we can safely return to last_win */
		pal->winvar = - pal->winvar;
	} else if( pal->winvar >= 0 ) { /* recently growing */
		pal->winvar = SCALE8(tp->snd_cwnd,win_small_decrease8);
	} else { /* recently decreasing */
		pal->winvar = SCALE8(tp->snd_cwnd,win_small_growth8);
	}
	pal->osc = 1;
}

/* During a windowful (i.e. refWinSegs) the window variation will be a winVar */
/* Anyhow make sure winVar variation is reached at least 2 acks before a windowful has passed */
static inline u32 calcNewWin(u32 refWinSegs, u32 deltaAckSegs, s32 winVar) {
	if( winVar < 0 )						return refWinSegs + winVar; /* no problem for fast decrease */
	if( deltaAckSegs < (u32)winVar )	return refWinSegs + deltaAckSegs; /* gradual increase */
	return refWinSegs + winVar; /* enough acks seen, no more growth needed */
}

#define SETOPTDATA { \
	pal->opt_rate = currentAckRate; \
	pal->opt_win = pal->last_win; \
	pal->opt_srtt_us = tp->srtt_us; }
	/* current rate is due to last_win, not current win ! */



#define aprox0(z8) ( -x_thresholdLOW8<z8 && z8<x_thresholdLOW8 )
#define aprox1(z8) ( (256-x_thresholdLOW8)<z8 && z8<(256+x_thresholdLOW8) )
#define above1(z8) ( z8>=x_thresholdL8 )			/* x_thresholdL8 is bigger than 256 */
#define below0(z8) ( z8<=(256-x_thresholdL8) )	/* x_thresholdL8 is bigger than 256 */
#define significant(z8) ( z8>=x_thresholdK8 )			/* x_thresholdK8 is smaller than 256 */
#define insigificant(z8) ( z8<=(256-x_thresholdK8) )	/* x_thresholdK8 is smaller than 256 */

#define capacityNotJointlyReached(z8)    ( useX ? aprox1(z8) : aprox0(z8) ) /* case A */
#define capacityExclusivelyTaken(z8)     ( useX ? aprox0(z8) : aprox1(z8) ) /* case B */
#define competitorsRetiringFromQueue(z8) ( useX ? above1(z8) : below0(z8) ) /* case C */
#define capacityJointlyTakenButHavingSmallShare(z8) ( useX ? significant(z8) : insigificant(z8) ) /* case A' */

/* This function is called after the following checks:
 *   We are limited by the congestion window (e.g. not application limited).
 *   At least a full window of acks have been transmitted while being limited by congestion win.
 *   The ack rate from the time of beginning of the last window of acks is known.
 *   Xa (the share of the available BW at the bottleneck is known.
 */
static inline void palermoWinVar(struct tcp_sock *tp, struct palermo *pal, s32 za8,u32 currentAckRate)
{
	if( pal->state >= queueRecover1State ) {
		/********** queueRecover1State or queueRecover2State *********************/
		/* recover from big opt_win (opt_rate overestimation is not covered yet) */
		if( currentAckRate < SCALE8(pal->opt_rate,rate_thresholdAPROX8) || tp->snd_cwnd<=2 )
		{ /* Below opt_rate: Finish queueRecover. */
			pal->state = regularState;
			pal->ls = 0; /* reset ls count */
			/* Always set new opt win here, this value may be suboptimum but safe */
			/* Current rate (below opt_rate) is due to last_win, so set opt_win to prev_last_win */
			pal->opt_win = tp->snd_ssthresh = pal->prev_last_win;
			setWinVarBounce(tp,pal);
			return;
		}
		/* Above or aprox opt_rate */
		if( currentAckRate > pal->opt_rate ) pal->opt_rate = currentAckRate; /* no opt_win setting here */
		/* continue queueRecover until we get below optRate */
		setWinVar(tp,pal,win_small_decrease8);
		return;
	}

	/* regularState, gentleFightState, or dogFight states  *******************/
	/* Check if xa8 =~ 0 (paper case B) */
	if( capacityExclusivelyTaken(za8) ) {
		if( currentAckRate > SCALE8(pal->opt_rate,rate_thresholdAPROX8) ) {
			/* Bottleneck capacity is taken exclusively by us ! */
			/* Check if this condition is repeating */
			if( pal->ls<0 ) { pal->ls = 1; } else { ++(pal->ls); }
			if( pal->ls>=x_REP ) {
				/* very low sensibility to cwnd size => opt_win may be overestimated */
				pal->state = queueRecover1State;
				if( currentAckRate > pal->opt_rate ) pal->opt_rate = currentAckRate; /* no opt rate setting here */
				pal->ls = 0; /* reset ls count */
				setWinVar(tp,pal,win_small_decrease8);
				return;
			} /* else maintain ls count */
		}
	} else {
		pal->ls = 0; /* reset ls count */
		/* Check if xa8 =~ 1 (paper case A) */
		if( capacityNotJointlyReached(za8) ) {
			/* In-flight size below the base bandwidth-delay-product and
		 	* bottleneck capacity no jointly taken by competing conns ! */
			/* perfect sensibility to cwnd size => go to regular growth! */
			if( currentAckRate > pal->opt_rate ) SETOPTDATA;
			/* should we always set opt data here? (in order to facilitate
				future growth instead of falling again to gentleFightState) */
			pal->state = regularState;
			setWinVar(tp,pal,win_regular_growth8);
			return;
		}
	}

	if( currentAckRate > SCALE8(pal->opt_rate,rate_thresholdAPROX8) ) {
		/********** at best performance or close to it ************************/
		s32 deltaRate    = dif32(currentAckRate,pal->last_ackrate);
		s32 optDeltaRate = dif32(currentAckRate,pal->opt_rate);
		s32 optDeltaWin;
		if( pal->state==regularState ) {
			/* Arrived here after regular win growth (inFlightData growth) */
			if( deltaRate>0 ) {
				if( optDeltaRate>0 ) SETOPTDATA;
				if( Z8_DEFINED(za8) ) {
					if( competitorsRetiringFromQueue(za8) ) { /* Xa very large */
						/* rate grew more than win, contributed by competition recess */
						setWinVar(tp,pal,win_small_growth8);
						return;
					} else if( capacityJointlyTakenButHavingSmallShare(za8) ) { /* Xa > 0.6 */
						/* rate grew according to win increase */
						setWinVar(tp,pal,win_regular_growth8);
						return;
					} /* else { } rate growth not significant */
				} else { /* za8 undefined, cannot use it here, so check deltaRate only */
			  		/* Rate grew with zero win growth, check if this growth is significant. */
					/* If significant, this again means competition recess, but in such case,
					 * we may still have to increase win to reach bdp */
					if( deltaRate > SCALE8(pal->last_ackrate,win_small_growth8) ) { 
						setWinVar(tp,pal,win_small_growth8);
						return;
					}
				}
			}
			/* Rate did not grew significantly, or decreased but not much. */
			setWinVarOscilate(tp,pal); /* to allow sensibility check */
			return;
		}

		/* state is gentleFight or dogFight  ****/
		if( optDeltaRate>0 ) {
			/* recalculate X^a from optimum values */
			/* (remember: do not alter opt data before comparisons with old opt data!! */
			int zo8 = useX ?	changeRatio8( currentAckRate,pal->opt_rate,    pal->last_win,pal->opt_win) : 
									changeRatio8( tp->srtt_us,   pal->opt_srtt_us, pal->last_win,pal->opt_win);
			if( Z8_DEFINED(zo8) ) {
				if( competitorsRetiringFromQueue(zo8) ) { /* Xo very large */
					/* currentRate grew much more than expected by pal->last_win value */
					pal->state = regularState;
					SETOPTDATA;
					setWinVar(tp,pal,win_small_growth8);
					return;
				} else if( capacityJointlyTakenButHavingSmallShare(zo8) ) { /* Xo > 0.6 */
					// currentRate grew in accordance to pal->last_win
					pal->state = regularState;
					SETOPTDATA;
					setWinVar(tp,pal,win_regular_growth8);
					return;
				} /* else { } rate growth not significant */
			} else { /* zo8 undefined, cannot use it here, so check deltaRate only */
			  	/* Rate grew with zero win growth, check if this growth is significant. */
				/* If significant, this again means competition recess, but in such case,
				 * we may still have to increase win to reach bdp */
				if( optDeltaRate > SCALE8(pal->opt_rate,win_small_growth8) ) { 
					pal->state = regularState;
					SETOPTDATA;
					setWinVar(tp,pal,win_small_growth8);
					return;
				}
			}
		}
		/* Rate did not grew significantly, or decreased but not much. */
		optDeltaWin = dif32(pal->last_win,pal->opt_win);

		if( optDeltaWin > SCALE8(pal->opt_win,win_regular_growth8) ) {
			/* pal->last_win grew too much for little rate gain */
			/* opt_win overestimated, go to recover window */
			pal->state = queueRecover2State;
			setWinVar(tp,pal,win_small_decrease8);
			return;
		}
		pal->state = regularState;

		/* pal->last_win already appropriate => leave everything unchanged but state */
		setWinVarOscilate(tp,pal) ; /* to allow sensibility check */
		return;
	}

	/******** Below best performance **************************/

	/* We are here because of competing connections at the bottleneck,
		or because of optRate is overestimated, or because we wrongly
		decreased congestion window too much.
		We do not assume sender app limitation because it prevents reaching this code.
		If competition is well behaved, it will notice it is not
		increasing rate as fast as it is increasing flightSize and will
		refrain from continue its growth without need for us to signal them.
		If it is bad behaved (regular TCP), it will continue its in-flight
		data growth and end pushing us to dogFight (regular RENO?).
		DogFight: we cannot differentiate between a single bad competitor
		and many gentle competitors. But in either case delay is already
		much grown and there is no gain in trying to prevent bufferbloat.
		*/
	if( (currentAckRate < SCALE8(pal->opt_rate,rate_thresholdDRS8) ) || ( pal->state == dogFightState && currentAckRate < (pal->opt_rate) >> 1))
	{ /* go to, or remain in dogFight (regular RENO/CUBIC!) */
		pal->state = dogFightState;
		tp->snd_ssthresh = (u32)(-1); /* reset ssthresh, will set again on winRecover or lossRecovery (fast or timeout). */
		/* no check currentrate vs last rate in this state!! always grow */
		setWinVarRegular(tp,pal);
		return;
	}
	/* CurrentRate below best, but not too much
	   It does not care last state, we go to gentleFight */
	if (pal->state != gentleFightState)
		pal->state = gentleFightState;
	setWinVarOscilate(tp,pal) ; /* to allow sensibility check */
}


#ifdef PALDEBUG

char* x8print(const char* str,int x8, char* buf) {
	int entero,frac; char *p=buf;
	*p++ = ' ';
	while( *str ) *p++ = *str++;
	if( x8<0 ) { *p++ = '-'; x8 = -x8; } 
	entero = x8>>8;
	frac = ((x8-(entero<<8))*1000)>>8;
	p += snprintf(p,31-(p-buf),"%d",entero);
	if( !frac ) return buf;
	p += snprintf(p,31-(p-buf),".%.3d",frac);
	--p;
	while( *p=='0' ) *p-- = 0;
	return buf;
}

#define DEBUGPRINFO(step,last_ack_us,deltaUs,last_ack,deltaAck,deltaAckSegs, \
					 	  currentAckRate,last_ackrate,deltaAckRate,last_win,prev_last_win) \
	if( *step=='3' ) \
		snprintf(buf,255," * xa8=%d ya8=%d %s %s",xa8,ya8,x8print("xa=",xa8,bufx),x8print("ya=",ya8,bufy)); \
	else *buf=0; \
	pr_info("tcp_cong_avoid(...) step=%s conn=%u-%pI4-%u * " \
		"nextState=%s us=%llu lastAckUs=%llu deltaUs=%llu * " \
		"ack=%u lastAck=%u deltaAck=%u deltaAckSegs=%u ackedSegs=%u * " \
		"currentAckKbps=%u lastAckKbps=%u deltaAckKbps=%d optAckKbps=%u * " \
		"win=%u newWin=%u deltaWin=%d winVar=%d lastWin=%u prevLastWin=%u optWin=%u ssthresh=%u * " \
		"maxPktsOut=%u cwndLimB=%hhu *" \
		"mss=%u advmss=%u * " \
		"rtt=%u rttdev=%u optRtt=%u * osc=%u %s", \
		step,htons(inet->inet_sport),&inet->inet_saddr,htons(inet->inet_dport), \
		pal_cstateTable[pal->state], \
		tp->tcp_mstamp, last_ack_us, deltaUs, \
		ack,last_ack,deltaAck,deltaAckSegs,ackedSegs, \
		currentAckRate<<3,last_ackrate<<3,deltaAckRate<<3,(pal->opt_rate)<<3, \
		tp->snd_cwnd,newcwnd,(newcwnd-tp->snd_cwnd),pal->winvar,last_win,prev_last_win,pal->opt_win,tp->snd_ssthresh, \
		tp->max_packets_out,tp->is_cwnd_limited, \
		tp->mss_cache,tp->advmss, \
		tp->srtt_us>>3,tp->mdev_us>>2,pal->opt_srtt_us>>3, \
		pal->osc,buf);

#else

#define DEBUGPRINFO(step,last_ack_us,deltaUs,last_ack,deltaAck,deltaAckSegs, \
					 	  currentAckRate,last_ackrate,deltaAckRate,last_win,prev_last_win)
#endif


/* this version is the same as tcp_is_cwnd_limited(sk) in include/net/tcp.h
 * except that it allows for approximate limitation by snd_cwnd */
static inline  bool pal_is_cwnd_limited(const struct sock *sk) {
	const struct tcp_sock *tp = tcp_sk(sk);
	/* If in slow start, ensure cwnd grows to twice what was ACKed. */
	if( tcp_in_slow_start(tp) ) return tp->max_packets_out > (tp->snd_cwnd>>1) ;
	if( ! tp->is_cwnd_limited ) return tp->max_packets_out > SCALE8(tp->snd_cwnd,rate_thresholdAPROX8);
   return (bool)1;
}

static void tcp_palermo_cong_avoid(struct sock *sk, u32 ack, u32 ackedSegs) {
	struct palermo *pal = inet_csk_ca(sk);
	struct tcp_sock *tp = tcp_sk(sk);
	struct inet_sock *inet = inet_sk(sk);
	u32 newcwnd;
	u32 deltaAck;     /* acked bytes since pal->last_ack, i.e. since pal->last_win was set */
	u32 deltaAckSegs; /* acked segments since pal->last_ack, i.e. when pal->last_win was set, */
	/* u32 ackedSegs:    acked segments since last ack received by stack: deltaAckSegs>=ackedSegs */
	u64 deltaUs;
	u32 currentAckRate;
	s32 deltaAckRate;
	s32 xa8,ya8;
	u64 aux;
#ifdef PALDEBUG
	char buf[256],bufx[64],bufy[64];
#endif

	if( !pal_is_cwnd_limited(sk) ) {
		/* not cwnd_limited
			=> cwnd no esta validada (ver RFC2861 y RFC7661, esta ultima no utilizada en linux)
			Al no controlar variaciones de inFlightSize ni pipeAck, entonces no podemos
			aplicar nuestro algoritmo de estimación del share del bottleneck.
			En el caso de RFC2861 por cada RTO en este estado se baja cwnd a (cwnd+cwnd_used)/2,
			pero se recuerda el cwnd incial en ssthresh para recuperar rapido si se vuelve cwnd_limited */
		pal->state = nonCwndLimState;
		pal->last_ack = 0; /* this suffices to reset algorithm */
#ifdef PALDEBUG
		pr_info("tcp_palermo_cong_avoid(...) step=dg not congestion limited! "
					"zeroing conn=%u-%pI4-%u * nextState=nonCwndLim * "
					"us=%llu win=%u ssthresh=%u * "
				   "maxPktsOut=%u cwndLimB=%hhu",
					htons(inet->inet_sport), &inet->inet_saddr, htons(inet->inet_dport),
					tp->tcp_mstamp,tp->snd_cwnd,tp->snd_ssthresh,
					tp->max_packets_out,tp->is_cwnd_limited);
#endif
		return;
	}

	if( !pal->last_ack ) { /* need to soft restart algorithm */
		/* No palermo state yet! First run after handshake or after cwr/recovery/loss.
		 * Anyhow we need to grow the window here, even when cwnd is not validated yet */
		setWinVarRegular(tp, pal);
		newcwnd = calcNewWin(tp->snd_cwnd,ackedSegs,pal->winvar);
		pal->state = initState;
		DEBUGPRINFO("1",0,0,0,0,0,0,0,0,0,0);
		/* reset pal data */
		pal->last_ack_us		= tp->tcp_mstamp;
		pal->last_ack			= ack;
		pal->last_ackrate		= 0;
		pal->last_srtt_us 	= tp->srtt_us;
		pal->prev_last_win	= 0;
		pal->last_win			= tp->snd_cwnd = newcwnd;
		pal->osc					= 0;
		pal->ls					= 0;
		pal->guard				= 0;
		/* Keep optimum values: pal->opt_win, pal->opt_rate, pal->opt_srtt_us
			Keep pal->winvar: already set in setWinVarRegular(tp,pal)~! */
		return;
	}
	/* tenemos last_ack y podemos calcular deltaAck u deltaUs */
	/* ( no hace falta usar  tcp_clock_ns() porque el timestamp del ack ya quedó en tp->tcp_mstamp ) */
	deltaAck			= dif32(ack,pal->last_ack); 
	deltaAckSegs	= deltaAck / tp->mss_cache;
	deltaUs			= dif32(tp->tcp_mstamp,pal->last_ack_us);

	if( pal->guard==0 ) { /* first window after checkpoint */ 
		if( deltaAck < (pal->last_win * tp->mss_cache) ) {
			/* Cannot measure xa yet. Because we still did not see a full window of acks. */
			/* Incrementally expand win. pal->winvar already calculated in last checkpoint */
			newcwnd = calcNewWin(pal->last_win,deltaAckSegs,pal->winvar);
			DEBUGPRINFO("partial",pal->last_ack_us,deltaUs,pal->last_ack,deltaAck,
							 deltaAckSegs,0,0,0,pal->last_win,pal->prev_last_win);
			tp->snd_cwnd = newcwnd;
			/* do not record anything here ! As we cannot measure xa, this is not a true checkpoint */
			return;
		}
		/* a first window sized as pal->last_win has passed */

		/* check pal->winvar was fully applied (should never be needed, but just in case) */
		u32 shouldWin = pal->last_win + pal->winvar;
		if( tp->snd_cwnd < shouldWin ) tp->snd_cwnd = shouldWin;

	} else {
		/* pal->guard > 0 
		 * keep tp->snd_cwnd constant during this interval */
		if( deltaAck < (tp->snd_cwnd * tp->mss_cache) ) {
			/* Cannot measure xa yet. Because we still did not see another full window of acks. */
			/* keep win and everything */
			return;
		}
		/* another window sized as tp->snd_cwnd has passed */
	}

	/* When reaching this point at least one full window has passed since pal->last_ack */
	/* If we have no previous recorded last_ackrate, we do not wait any guard windows, 
	 * But in regular operations (i.e. having last_ackrate) we wait some guardWindows before
	 * measuring Xa */
	if( pal->last_ackrate && pal->guard<guardWindows ) {
		DEBUGPRINFO("guard",pal->last_ack_us,deltaUs,pal->last_ack,deltaAck,
						deltaAckSegs,0,0,0,pal->last_win,pal->prev_last_win);
		/* reset last_ack and last_ack_us in order to restart clean measurement of last_win */
		pal->last_ack_us	= tp->tcp_mstamp;
		pal->last_ack		= ack;
		if( pal->guard==0 ) {
			if( pal->state!=initState ) {
				/* When in initState, last_ackrate was not due to last_last_win, but to prev_last_win */
				/* because of step 2 skipping guardWindow (a little dirty but faster) */
				pal->prev_last_win = pal->last_win; 
			}
			pal->last_win = tp->snd_cwnd;
		}
		++(pal->guard);
		return;
	}
	pal->guard = 0;

	/* A full window of acks was received after other full 'guardWindows' have passed */

	/* Let's measure xa8. Checkpoint */
	aux = (u64)deltaAck; aux *= 1000; aux /= deltaUs;
	currentAckRate = (u32)aux; /* kilobytes per second*/
	if( ! pal->last_ackrate ) {
		/* We still do not have last_ackrate, so we cannot calc deltaAckRate nor xa
		 * We still need to calculate a new win though */
		setWinVarRegular(tp, pal);
		newcwnd = calcNewWin(tp->snd_cwnd,ackedSegs,pal->winvar);
		/* check also if we still do not have an opt rate and win (i.e. very beggining of conn) */
		if( pal->opt_rate==0 ) SETOPTDATA;
		DEBUGPRINFO("2",pal->last_ack_us,deltaUs,pal->last_ack,deltaAck,
						deltaAckSegs,currentAckRate,0,0,pal->last_win,pal->prev_last_win);
		goto savePalData;
	}

	deltaAckRate = dif32(currentAckRate,pal->last_ackrate);
	xa8 = changeRatio8( currentAckRate,pal->last_ackrate, pal->last_win,pal->prev_last_win);
	ya8 = changeRatio8( tp->srtt_us,pal->last_srtt_us, pal->last_win,pal->prev_last_win);
	palermoWinVar(tp, pal, (useX?xa8:ya8), currentAckRate);
	newcwnd = calcNewWin(tp->snd_cwnd,ackedSegs,pal->winvar);
	DEBUGPRINFO("3",pal->last_ack_us,deltaUs,pal->last_ack,deltaAck,deltaAckSegs,
						 currentAckRate,pal->last_ackrate,deltaAckRate,pal->last_win,pal->prev_last_win);

savePalData:
	pal->last_ack_us	 = tp->tcp_mstamp;
	pal->last_ack		 = ack;
	pal->last_ackrate  = currentAckRate;
	pal->last_srtt_us  = tp->srtt_us;
	pal->prev_last_win = pal->last_win;
	pal->last_win      = tp->snd_cwnd;
	tp->snd_cwnd		 = newcwnd;
}


/* new value of cwnd after loss (required) */
/* Es llamada para que calcule el nuevo valor de tp->snd_cwnd cuando
	corresponde deshacer una reducción de la ventana,
	es decir que se resolvió la causa que la había bajado.
	En el caso de reno devuelve simplemente max(tp->snd_cwnd, tp->prior_cwnd).
	(El stack de tcp también hace tp->snd_ssthresh=tp->prior_ssthresh
	sin consultar al módulo de control de congestión)
 */
static u32 tcp_palermo_undo_cwnd(struct sock *sk)
{
	const struct tcp_sock *tp = tcp_sk(sk);
	struct palermo *pal = inet_csk_ca(sk);
	struct inet_sock *inet = inet_sk(sk);
	
#ifdef PALDEBUG
	pr_info("tcp_palermo_undo_cwnd() step=undoCwnd conn=%u-%pI4-%u * ",
	htons(inet->inet_sport),
	&inet->inet_saddr,
	htons(inet->inet_dport));
#endif

	pal->last_ack = 0; /* this suffices to reset algorithm */
	return max(tp->snd_cwnd, tp->prior_cwnd);
}

/* Es llamada cada vez que se cambia el estado, para que el algoritmo de control de
 * congestión setee también su propio estado */
void tcp_palermo_set_state(struct sock *sk, u8 new_state) {
	struct inet_sock *inet = inet_sk(sk);
	switch(new_state) {
		case TCP_CA_Open:	  /* (0) Todo bien */
#ifdef PALDEBUG
			pr_info("tcp_palermo_set_state(CA_Open) conn=%u-%pI4-%u * ",
			htons(inet->inet_sport),
			&inet->inet_saddr,
			htons(inet->inet_dport));
#endif
			break;
		case TCP_CA_Disorder: /* (1) Received reordered acks, but not enough to assume loss */
#ifdef PALDEBUG
			pr_info("tcp_palermo_set_state(CA_Disorder) conn=%u-%pI4-%u * ",
			htons(inet->inet_sport),
			&inet->inet_saddr,
			htons(inet->inet_dport));
#endif
			break;
		case TCP_CA_CWR:		/* (2) Congestion Window Reduction (CWR) state: received ACKs with ECN-ECE marks, or problems in output queue (but not in the net) */
#ifdef PALDEBUG
			pr_info("tcp_palermo_set_state(CA_CWR) pal last ack reset, conn=%u-%pI4-%u * ",
			htons(inet->inet_sport),
			&inet->inet_saddr,
			htons(inet->inet_dport));
#endif
			((struct palermo *)inet_csk_ca(sk))->last_ack = 0; /* this suffices to reset algorithm */
			break;
		case TCP_CA_Recovery: /* (3) In fast recovery and retransmitting lost packets, typically triggered by ACK events. */
#ifdef PALDEBUG
			pr_info("tcp_palermo_set_state(CA_Recovery) pal last ack reset for conn=%u-%pI4-%u * ",
			htons(inet->inet_sport),
			&inet->inet_saddr,
			htons(inet->inet_dport));
#endif
			((struct palermo *)inet_csk_ca(sk))->last_ack = 0; /* this suffices to reset algorithm */
			break;
		case TCP_CA_Loss:	  /* (4) in loss recovery triggered by retransmission timeout. */
#ifdef PALDEBUG
			pr_info("tcp_palermo_set_state(CA_Loss) pal last ack reset for conn=%u-%pI4-%u * ",
			htons(inet->inet_sport),
			&inet->inet_saddr,
			htons(inet->inet_dport));
#endif
			((struct palermo *)inet_csk_ca(sk))->last_ack = 0; /* this suffices to reset algorithm */
			break;
		default:
#ifdef PALDEBUG
			pr_info("tcp_palermo_set_state(CA_Unknown) conn=%u-%pI4-%u * ",
			htons(inet->inet_sport),
			&inet->inet_saddr,
			htons(inet->inet_dport));
#endif
	}
}

/* es llamado cuando ocurre algun evento de control de congestión */
static void tcp_palermo_event(struct sock *sk, enum tcp_ca_event event)	{
	
	struct inet_sock *inet = inet_sk(sk);
	
	switch (event) {
	case CA_EVENT_COMPLETE_CWR: /* end of congestion recovery */
#ifdef PALDEBUG
		pr_info("tcp_palermo_event(CA_EVENT_COMPLETE_CWR) conn=%u-%pI4-%u * ",
			htons(inet->inet_sport),
			&inet->inet_saddr,
			htons(inet->inet_dport));
#endif
		break;
	case CA_EVENT_LOSS:			/* loss timeout */
#ifdef PALDEBUG
          pr_info("tcp_palermo_event(CA_EVENT_LOSS) conn=%u-%pI4-%u * ",
                htons(inet->inet_sport), &inet->inet_saddr,
                htons(inet->inet_dport));
#endif
        
		break;
	case CA_EVENT_ECN_IS_CE:	 /* received CE marked IP packet */
#ifdef PALDEBUG
		pr_info("tcp_palermo_event(CA_EVENT_ECN_IS_CE) conn=%u-%pI4-%u * ",
			htons(inet->inet_sport),
			&inet->inet_saddr,
			htons(inet->inet_dport));
#endif
		break;
	case CA_EVENT_ECN_NO_CE:	 /* ECT set, but not CE marked */
#ifdef PALDEBUG
		pr_info("tcp_palermo_event(CA_EVENT_ECN_NO_CE) conn=%u-%pI4-%u * ",
			htons(inet->inet_sport),
			&inet->inet_saddr,
			htons(inet->inet_dport));
#endif
		break;
	case CA_EVENT_CWND_RESTART: /* congestion window restart */
#ifdef PALDEBUG
        pr_info("tcp_palermo_event(CA_EVENT_CWND_RESTART) conn="
                  "%u-%pI4-%u * ( idleTime>N*RTO => cwnd/= 2^N )",
                htons(inet->inet_sport), &inet->inet_saddr,
                htons(inet->inet_dport));
#endif
        break;
	case CA_EVENT_TX_START:	  /* first transmit when no packets in flight */
#ifdef PALDEBUG
		pr_info("tcp_palermo_event(CA_EVENT_TX_START) conn=%u-%pI4-%u * ",
			htons(inet->inet_sport),
			&inet->inet_saddr,
			htons(inet->inet_dport));
#endif
		break;
	default:
#ifdef PALDEBUG
		pr_info("tcp_palermo_event(CA_UNKNOWN_EVENT conn=%u-%pI4-%u * ",
			htons(inet->inet_sport),
			&inet->inet_saddr,
			htons(inet->inet_dport));
#endif
		break;
	}
}

/* return slow start threshold (required) */
/* la recalcula, en el caso de reno, la baja a la mitad.
 * Se llama cuando se detectó una pérdida */
u32 tcp_palermo_ssthresh(struct sock *sk)
{
	const struct tcp_sock *tp = tcp_sk(sk);
	struct inet_sock *inet = inet_sk(sk);
#ifdef PALDEBUG
	pr_info("tcp_palermo_ssthresh() step=ssthreshDown conn=%u-%pI4-%u",
								htons(inet->inet_sport),
								&inet->inet_saddr,
								htons(inet->inet_dport));
#endif
	return max((3*tp->snd_cwnd) >> 2U, 2U);
}

static struct tcp_congestion_ops tcp_palermo __read_mostly = {
	.init		= tcp_palermo_init,
	.ssthresh	= tcp_palermo_ssthresh,
	.cong_avoid	= tcp_palermo_cong_avoid,
	.set_state  = tcp_palermo_set_state,
	.cwnd_event	= tcp_palermo_event,
	.undo_cwnd	= tcp_palermo_undo_cwnd,
	.owner		= THIS_MODULE,
	.name		= "palermo"
};

static int __init tcp_palermo_register(void)
{
	BUILD_BUG_ON(sizeof(struct palermo) > ICSK_CA_PRIV_SIZE);
	return tcp_register_congestion_control(&tcp_palermo);
}

static void __exit tcp_palermo_unregister(void)
{
	tcp_unregister_congestion_control(&tcp_palermo);
}

module_init(tcp_palermo_register);
module_exit(tcp_palermo_unregister);

MODULE_AUTHOR("Alejandro Popovsky, Gustavo Muzzillo");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("TCP Palermo+");
